#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   Z80 CPU Emulator.
#   https://github.com/kosarev/z80
#
#   Copyright (C) 2017-2020 Ivan Kosarev.
#   ivan@kosarev.info
#
#   Published under the MIT license.

import bisect
from ._error import Error


class _SourcePos(object):
    def __init__(self, offset, source_file):
        self.__offset = offset
        self.__source_file = source_file

    def __repr__(self):
        file = self.__source_file
        line, line_no, column_no = file.get_coordinates(self.__offset)
        return '%s\n%s^\n%s:%s:%s' % (line, ' ' * column_no, file,
                                      line_no, column_no)


class _SourceFile(object):
    def __init__(self, filename):
        self.__filename = filename

        with open(self.__filename) as f:
            self.__image = f.read()

        self.__line_breaks = tuple(
            i for i, c in enumerate(self.__image) if c == '\n')

    def __repr__(self):
        return self.__filename

    def get_image(self):
        return self.__image

    def get_coordinates(self, offset):
        i = bisect.bisect_left(self.__line_breaks, offset)

        line_start = self.__line_breaks[i - 1] + 1 if i > 0 else 0
        line_end = (self.__line_breaks[i] if i < len(self.__line_breaks)
                    else len(self.__image))
        line = self.__image[line_start:line_end]

        line_no = i + 1
        column_no = offset - line_start

        return line, line_no, column_no


class _SourceError(Error):
    def __init__(self, subject, message):
        super().__init__('%r: %s' % (subject, message))


class _Tokenizer(object):
    def __init__(self, source_file):
        self.__source_file = source_file
        self.__image = self.__source_file.get_image()
        self.__offset = 0
        self.__token_offset = self.__offset

    def __get_end_offset(self):
        return len(self.__image)

    def __find_all(self, *anchors):
        # Remember the offset to not depend on its changes.
        start_offset = self.__offset
        for a in anchors:
            i = self.__image.find(a, start_offset)
            if i >= 0:
                yield a, i

    def __find_next(self, *anchors):
        anchor, offset = None, None
        for a, i in self.__find_all(*anchors):
            if offset is None or offset > i:
                anchor, offset = a, i
        return anchor, offset

    def skip_to(self, *anchors):
        anchor, offset = self.__find_next(*anchors)
        if offset is None:
            self.__offset = self.__get_end_offset()
        else:
            self.__offset = offset
        return anchor

    def skip_rest_of_line(self):
        self.skip_to('\n')

    def skip_next(self, *anchors):
        anchor = self.skip_to(*anchors)
        if anchor is not None:
            self.__offset += len(anchor)
        return anchor

    def __get_front(self, size=None):
        end = self.__offset + size if size is not None else None
        return self.__image[self.__offset:end]

    def __follows_with(self, *fillers):
        for filler in fillers:
            if self.__get_front().startswith(filler):
                return filler
        return None

    def __skip(self, *fillers):
        while True:
            filler = self.__follows_with(*fillers)
            if filler is None:
                break

            self.__offset += len(filler)

    def skip_whitespace(self):
        self.__skip(' ', '\t')

    def skip_char(self):
        c = self.__get_front(1)
        self.__offset += 1
        return c

    def start_token(self):
        self.__token_offset = self.__offset

    def end_token(self):
        pos = _SourcePos(self.__token_offset, self.__source_file)

        literal = None
        if self.__token_offset < self.__offset:
            literal = self.__image[self.__token_offset:self.__offset]

        return literal, pos


class _ProfileTag(object):
    def __init__(self, kind, addr, comment):
        self._kind = kind
        self._addr = addr
        self._comment = comment

    def get_kind(self):
        return self._kind

    def get_addr(self):
        return self._addr

    def get_comment(self):
        return self._comment

    def __repr__(self):
        comment = self.get_comment()
        comment = '' if not comment else ' : %s' % comment
        return '@@ 0x%04x %s%s' % (self.get_addr(), self.get_kind(), comment)


class _InstrTag(_ProfileTag):
    ID = 'instr'

    def __init__(self, addr, comment):
        super().__init__(self.ID, addr, comment)


class _Token(object):
    def __init__(self, literal, source_pos):
        self.__literal = literal
        self.__source_pos = source_pos

    def __repr__(self):
        return '%r: %r' % (self.__source_pos, self.__literal)

    def get_literal(self):
        return self.__literal

    def get_source_pos(self):
        return self.__source_pos


class _TagParser(object):
    __DELIMITERS = '\n :,()[]'

    def __init__(self, source_file):
        self.__toks = _Tokenizer(source_file)

    def __fetch_token(self, error=None):
        self.__toks.skip_whitespace()

        self.__toks.start_token()
        if self.__toks.skip_char() not in self.__DELIMITERS:
            self.__toks.skip_to(*self.__DELIMITERS)
        literal, pos = self.__toks.end_token()

        if literal == '\n':
            literal = None

        if literal is not None:
            return _Token(literal, pos)

        if error is not None:
            raise _SourceError(pos, error)

        return None

    def __parse_tag_address(self):
        tok = self.__fetch_token('Tag address expected.')
        try:
            return int(tok.get_literal(), base=0)
        except ValueError:
            raise _SourceError(tok, 'Malformed tag address.')

    def __parse_tag_name(self):
        return self.__fetch_token('Tag name expected.')

    def __parse_optional_comment(self):
        tok = self.__fetch_token()
        if tok is None:
            return None

        if tok.get_literal() == ':':
            self.__toks.skip_whitespace()
            self.__toks.start_token()
            self.__toks.skip_rest_of_line()
            literal, pos = self.__toks.end_token()
            return literal

        raise _SourceError(tok, 'End of line or a comment expected.')

    def __parse_instr_tag(self, addr, name):
        comment = self.__parse_optional_comment()
        return _InstrTag(addr, comment)

    __TAG_PARSERS = {
        _InstrTag.ID: __parse_instr_tag,
    }

    # Parses and returns a subsequent tag.
    def __iter__(self):
        while self.__toks.skip_next('@@'):
            addr = self.__parse_tag_address()
            name = self.__parse_tag_name()

            parser = self.__TAG_PARSERS.get(name.get_literal(), None)
            if not parser:
                raise _SourceError(name, 'Unknown tag.')

            yield parser(self, addr, name)


class _Profile(object):
    __tags = dict()

    def load_if_exists(self, filename):
        try:
            file = _SourceFile(filename)
        except FileNotFoundError:
            return

        for tag in _TagParser(file):
            self.__tags.setdefault(tag.get_addr(), []).append(tag)


class _Disassembler(object):
    def __init__(self, image):
        self.__image = image

    def disassemble(self, profile):
        self._tags_to_process = []
        assert 0, profile._Profile__tags




#   Z80 CPU Emulator.
#   https://github.com/kosarev/z80
#
#   Copyright (C) 2017-2019 Ivan Kosarev.
#   ivan@kosarev.info
#
#   Published under the MIT license.

from ._machine import I8080Machine
from ._machine import Z80Machine
from ._main import main



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   Z80 CPU Emulator.
#   https://github.com/kosarev/z80
#
#   Copyright (C) 2017-2020 Ivan Kosarev.
#   ivan@kosarev.info
#
#   Published under the MIT license.

import sys
from ._disassembler import _Profile, _Disassembler
from ._error import Error


def _pop_argument(args, error):
    if not args:
        raise Error(error)

    return args.pop(0)


def _handle_extra_arguments(args):
    if args:
        raise Error('Extra argument %r.' % args[0])


def _disasm(args):
    image_filename = _pop_argument(args, 'The image file to disassemble is '
                                         'not specified.')
    asm_filename = _pop_argument(args, 'The assembly file is not specified.')
    profile_filename = _pop_argument(args, 'The profile is not specified.')
    _handle_extra_arguments(args)

    with open(image_filename, 'rb') as f:
        image = f.read()

    IMAGE_SIZE = 0x10000
    if len(image) != IMAGE_SIZE:
        raise Error('The image file shall be of exactly %d bytes in '
                    'size.' % IMAGE_SIZE)

    profile = _Profile()
    profile.load_if_exists(asm_filename)
    profile.load_if_exists(profile_filename)

    d = _Disassembler(image)
    d.disassemble(profile)
    assert 0, d  # TODO


def _handle_command_line(args):
    if not args:
        raise Error('Nothing to do.')

    command = args.pop(0)
    if command == 'disasm':
        _disasm(args)
        return

    raise Error('Unknown command %r.' % command)


def main():
    try:
        _handle_command_line(sys.argv[1:])
    except Error as e:
        lines = e.args[0].split('\n')
        lines[-1] = 'z80: %s' % lines[-1]
        sys.exit('\n'.join(lines))


if __name__ == "__main__":
    main()



#   Z80 CPU Emulator.
#   https://github.com/kosarev/z80
#
#   Copyright (C) 2017-2019 Ivan Kosarev.
#   ivan@kosarev.info
#
#   Published under the MIT license.


class Error(Exception):
    def __init__(self, reason):
        super().__init__(reason)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import distutils.sysconfig
import inspect
import os
# from distutils.core import Extension
from setuptools import Extension, setup


here = os.path.abspath(os.path.dirname(inspect.getsource(lambda: 0)))

with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


# Work around the problem with the warning about '-Wstrict-prototypes'.
# https://bugs.python.org/issue1222585
config_vars = distutils.sysconfig.get_config_vars()
opt_to_remove = '-Wstrict-prototypes'
for var in ['OPT']:
    if var in config_vars:
        opts = config_vars[var].split()
        if opt_to_remove in opts:
            opts.remove(opt_to_remove)
    config_vars[var] = ' '.join(opts)


z80_emulator_module = Extension(
    name='z80._z80',
    extra_compile_args=['-std=c++11', '-Wall', '-fno-exceptions', '-fno-rtti',
                        '-O3',
                        # '-S', '-fverbose-asm',  # TODO
                        ],
    sources=['z80/_z80module.cpp'],
    language='c++')


# TODO: Update the URL once we have a published documentation.
# TODO: Do we have a name for the emulator?
setup(name='z80',
      version='0.1a2',
      description='Fast and flexible i8080/Z80 Emulator',
      long_description=long_description,
      long_description_content_type='text/markdown',
      author='Ivan Kosarev',
      author_email='ivan@kosarev.info',
      url='https://github.com/kosarev/z80',
      license='MIT',
      ext_modules=[z80_emulator_module],
      packages=['z80'],
      install_requires=[],
      entry_points={
          'console_scripts': [
              'z80 = z80:main',
          ],
      },
      # TODO: test_suite='test',
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Intended Audience :: Developers',
          'Intended Audience :: Information Technology',
          'Intended Audience :: Education',
          'License :: OSI Approved :: MIT License',
          'Operating System :: OS Independent',
          'Programming Language :: C++',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: Implementation :: CPython',
          'Topic :: Software Development',
          'Topic :: Software Development :: Libraries',
          'Topic :: System :: Emulators',
      ],
      )




#   Z80 CPU Emulator.
#   https://github.com/kosarev/z80
#
#   Copyright (C) 2017-2019 Ivan Kosarev.
#   ivan@kosarev.info
#
#   Published under the MIT license.

import struct
from ._z80 import _I8080Machine, _Z80Machine


class _StateBase(object):
    _STATE_FIELDS = {
        'c': (0, 'B'), 'b': (1, 'B'), 'bc': (0, '<H'),
        'e': (2, 'B'), 'd': (3, 'B'), 'de': (2, '<H'),
        'l': (4, 'B'), 'h': (5, 'B'), 'hl': (4, '<H'),
        'f': (2, 'B'), 'a': (7, 'B'), 'af': (6, '<H'),
        'pc': (8, '<H'), 'sp': (10, '<H'),
    }

    def __init__(self, image):
        self._image = image

    def get(self, id):
        offset, format = self._STATE_FIELDS[id]
        size = struct.calcsize(format)
        return struct.unpack(format, self._image[offset:offset + size])[0]

    def set(self, id, n):
        offset, format = self._STATE_FIELDS[id]
        size = struct.calcsize(format)
        self._image[offset:offset + size] = struct.pack(format, n)

    def get_c(self):
        return self.get('c')

    def get_e(self):
        return self.get('e')

    def get_bc(self):
        return self.get('bc')

    def get_de(self):
        return self.get('de')

    def get_pc(self):
        return self.get('pc')

    def set_pc(self, n):
        self.set('pc', n)

    def get_memory_byte(self, addr):
        memory = self._get_memory_view()
        return memory[addr]

    def set_memory_block(self, addr, block):
        memory = self._get_memory_view()
        memory[addr:addr + len(block)] = block


class I8080State(_StateBase):
    def __init__(self, image):
        _StateBase.__init__(self, image)
        self._STATE_FIELDS.update({
            'iff': (16, 'B'), 'int_disabled': (17, 'B'), 'halted': (18, 'B'),
        })

    def _get_memory_view(self):
        return self._image[20:]


class Z80State(_StateBase):
    def __init__(self, image):
        super().__init__(image)
        self._STATE_FIELDS.update({
            'ixl': (16, 'B'), 'ixh': (17, 'B'), 'ix': (16, '<H'),
            'iyl': (18, 'B'), 'iyh': (19, 'B'), 'iy': (18, '<H'),

            'alt_c': (20, 'B'), 'alt_b': (21, 'B'), 'alt_bc': (20, '<H'),
            'alt_e': (22, 'B'), 'alt_d': (23, 'B'), 'alt_de': (22, '<H'),
            'alt_l': (24, 'B'), 'alt_h': (25, 'B'), 'alt_hl': (24, '<H'),
            'alt_f': (26, 'B'), 'alt_a': (27, 'B'), 'alt_af': (26, '<H'),

            'r': (28, 'B'), 'i': (29, 'B'), 'iy': (28, '<H'),
            'iff1': (30, 'B'), 'iff2': (31, 'B'),

            'int_disabled': (32, 'B'), 'halted': (33, 'B'),
            'int_mode': (34, 'B'), 'index_rp_kind': (35, 'B'),
        })

    def _get_memory_view(self):
        return self._image[36:]


class _MachineBase(object):
    # Events.
    _NO_EVENTS = 0
    _END_OF_FRAME = 1 << 0
    _BREAKPOINT_HIT = 1 << 1

    # Address marks.
    _NO_MARKS = 0
    _BREAKPOINT_MARK = 1 << 0

    def mark_addr(self, addr, marks):
        self.mark_addrs(addr, 1, marks)

    def set_breakpoint(self, addr):
        self.mark_addr(addr, self._BREAKPOINT_MARK)


class I8080Machine(_MachineBase, _I8080Machine, I8080State):
    def __init__(self):
        I8080State.__init__(self, self.get_state_view())


class Z80Machine(_MachineBase, _Z80Machine, Z80State):
    def __init__(self):
        Z80State.__init__(self, self.get_state_view())
